#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#----------------------------------------------------------------------
# xoutil.modules
#----------------------------------------------------------------------
# Copyright (c) 2013, 2014 Merchise Autrement
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License (GPL) as published by the
# Free Software Foundation;  either version 2  of  the  License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#
# Created on 13 janv. 2013

'''Modules utilities.
'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_import)

from types import ModuleType


def force_module(ref=None):
    '''Load a module from a string or return module if already created.

    If `ref` is not specified (or integer) calling module is assumed looking
    in the stack.

    .. note:: Implementation detail

       Function used to inspect the stack is not guaranteed to exist in all
       implementations of Python.

    '''
    if isinstance(ref, ModuleType):
        return ref
    else:
        if ref is None:
            ref = 1
        if isinstance(ref, int):
            import sys
            ref = sys._getframe(ref).f_globals['__name__']
        if not isinstance(ref, str):
            if isinstance(ref, bytes):
                ref = ref.decode()  # Python 3.x
            else:
                try:
                    ref = ref.encode()  # Python 2.x
                except:
                    msg = ("invalid type '%s' for module name '%s'" %
                           (type(ref), ref))
                    raise TypeError(msg)
        return __import__(ref, fromlist=[ref], level=0)


def copy_members(source=None, target=None):
    '''Copy module members from `source` to `target`.

    It's common in `xoutil` package to extend Python modules with the same
    name, for example `xoutil.datetime` has all public members of Python's
    `datetime`.  :func:`copy_members` can be used to copy all members from the
    original module to the extended one.

    :param source: string with source module name or module itself.

        If not given, is assumed as the last module part name of `target`.

    :param target: string with target module name or module itself.

        If not given, target name is looked in the stack of caller module.

    :returns: Source module.
    :rtype: `ModuleType`

    .. warning:: Implementation detail

       Function used to inspect the stack is not guaranteed to exist in all
       implementations of Python.

    '''
    target = force_module(target or 2)
    if source is None:
        source = target.__name__.rsplit('.')[-1]
        if source == target.__name__:
            msg = '"source" and "target" modules must be different.'
            raise ValueError(msg)
    source = force_module(source)
    for attr in dir(source):
        if not attr.startswith('__'):
            setattr(target, attr, getattr(source, attr))
    return source


class _CustomModuleBase(ModuleType):
    pass


def customize(module, custom_attrs=None, meta=None):
    '''Replaces a `module` by a custom one.

    Injects all kwargs into the newly created module's class. This allows to
    have module into which we may have properties or other type of
    descriptors.

    :param module: The module object to customize.

    :param custom_attrs: A dictionary of custom attributes that should be
        injected in the customized module.

        .. versionadded:: 1.4.2 Changes the API, no longer uses the
                          ``**kwargs`` idiom for custom attributes.

    :param meta: The metaclass of the module type. This should be a subclass
                 of `type`. We will actually subclass this metaclass to
                 properly inject `custom_attrs` in our own internal
                 metaclass.

    :returns: A tuple of ``(module, customized, class)`` with the module in
              the first place, `customized` will be True only if the module
              was created (i.e :func:`customize` is idempotent), and the
              third item will be the class of the module (the first item).

    '''
    if not isinstance(module, _CustomModuleBase):
        import sys
        from xoutil.objects import metaclass
        meta_base = meta if meta else type

        class CustomModuleType(meta_base):
            def __new__(cls, name, bases, attrs):
                if custom_attrs:
                    attrs.update(custom_attrs)
                return super(CustomModuleType, cls).__new__(cls, name, bases,
                                                            attrs)

        class CustomModule(metaclass(CustomModuleType), _CustomModuleBase):
            def __getattr__(self, attr):
                self.__dict__[attr] = result = getattr(module, attr)
                return result

            def __dir__(self):
                res = set(dir(module))
                if custom_attrs:
                    res |= set(custom_attrs.keys())
                return list(res)

        sys.modules[module.__name__] = result = CustomModule(module.__name__)
        return result, True, CustomModule
    else:
        return module, False, type(module)


def modulemethod(func):
    '''Decorator that defines a module-level method.

    Simply a module-level method, will always receive a first argument `self`
    with the module object.

    '''
    import sys
    from functools import wraps
    self, _created, cls = customize(sys.modules[func.__module__])
    @wraps(func)
    def inner(*args, **kwargs):
        return func(self, *args, **kwargs)
    setattr(cls, func.__name__, func)
    return inner


def moduleproperty(getter, setter=None, deleter=None, doc=None):
    '''Decorator that creates a module-level property.

    The module of the `getter` is replaced by a custom implementation of the
    module, and the property is injected to the custom module's class.

    '''
    import sys
    module = sys.modules[getter.__module__]
    module, _created, cls = customize(module)
    class prop(property):
        def setter(self, func):
            result = super(prop, self).setter(func)
            setattr(cls, func.__name__, result)
            return result

        def deleter(self, func):
            result = super(prop, self).deleter(func)
            setattr(cls, func.__name__, result)
            return result

    result = prop(getter, setter, deleter, doc)
    setattr(cls, getter.__name__, result)
    return result


def get_module_path(module):
    '''Gets the absolute path of a `module`.

    :param module: Either module object or a (dotted) string for the module.

    :returns: The path of the module.

    If the module is a package, returns the directory path (not the path to the
    ``__init__``).

    If `module` is a string and it's not absolute, raises a TypeError.

    '''
    from importlib import import_module
    from xoutil.fs.path import normalize_path
    from xoutil.six import string_types as strs
    mod = import_module(module) if isinstance(module, strs) else module
    # The __path__ only exists for packages and does not include the
    # __init__.py
    path = mod.__path__[0] if hasattr(mod, '__path__') else mod.__file__
    return normalize_path(path)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#----------------------------------------------------------------------
# xoutil.tests.test_aop
#----------------------------------------------------------------------
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# Copyright (c) 2012 Medardo Rodríguez
# All rights reserved.
#
# Contributors: see CONTRIBUTORS and HISTORY file
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on Mar 26, 2012

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode)


import unittest

from datetime import timedelta
from xoutil.aop import weaved
from xoutil.aop.basic import contextualized, complementor
from xoutil.six.moves import range


range_ = lambda *args: list(range(*args))


def days(self):
    return timedelta(days=self)


def years(self):
    return timedelta(days=365 * self)


def months(self):
    'Regards a month with 30 days'
    return timedelta(days=30 * self)


class Test(unittest.TestCase):
    def setUp(self):
        class SimpleObject(object):
            def ident(self, what):
                return what

        self.SimpleObject = SimpleObject

    def test_simple_substitution(self):
        def plusone(self, what):
            return super(_class, self).ident(what) + 1

        sobj = self.SimpleObject()
        prev = sobj.ident(99)
        with weaved(sobj, ident=plusone) as sobj:
            _class = sobj.__class__
            self.assertEqual(prev + 1, sobj.ident(99))

    def test_nested_weaved(self):
        def plusone(self, what):
            return super(_class, self).ident(what) + 1

        def plustwo(self, what):
            return super(_class2, self).ident(what) + 2

        sobj = self.SimpleObject()
        prev = sobj.ident(99)
        with weaved(sobj, ident=plusone) as sobj:
            _class = sobj.__class__
            with weaved(sobj, ident=plustwo) as sobj:
                _class2 = sobj.__class__
                self.assertEqual(prev + 3, sobj.ident(99))
            self.assertEqual(prev + 1, sobj.ident(99))
        self.assertEqual(prev, sobj.ident(99))

    def test_complementor(self):
        def __init__(self, *args, **kw):
            return self._super___init__(*args, **kw)

        def other(self):
            'Hacked'
            return self._super_other()

        class Complementor(object):
            somelist = range_(10, 20)

            def other2(self):
                return 'other'

            @classmethod
            def clmethod(cls):
                return cls

            @staticmethod
            def stmethod():
                return 1

        @complementor(other, __init__, Complementor,
                      somedict={'a': 1, 'b': 2}, somelist=range_(5, 10))
        class Someclass(object):
            somedict = {'a': 0}
            somelist = range_(5)

            def __init__(self, d=None, l=None):
                'My docstring'
                if d:
                    self.somedict.update(d)
                if l:
                    self.somelist.extend(l)

            def other(self):
                'Other docstring'
                return self

        ok = self.assertEqual
        ok(Someclass.somedict, {'a': 1, 'b': 2})
        ok(Someclass.somelist, range_(5) + range_(10, 20))
        ok(Someclass.__init__.__doc__, 'My docstring')
        ok(Someclass.other.__doc__, 'Hacked')
        inst = Someclass()
        ok(inst.other2(), 'other')
        ok(Someclass.stmethod(), 1)
        ok(Someclass.clmethod(), Someclass)

    def test_contextualizer(self):
        from xoutil.context import context

        class FooBazer(object):
            def inside(self):
                if context['in-context']:
                    return 'in-context'
                else:
                    return 'KABOOM'

        @contextualized(context('in-context'), FooBazer)
        class Foobar(object):
            def outside(self):
                return 'not-contextualized'

        foo = Foobar()
        self.assertEqual('in-context', foo.inside())


if __name__ == "__main__":
    unittest.main(verbosity=2)



# Back-ported from Python 3.2.
# Copyright (c) 2001-2012, 2014, Python Software Foundation; All rights reserved.
# Retains the licence of the Python Software Foundation.

"""Redo the builtin repr() (representation) but with limits on most sizes."""


from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import)

import sys

if sys.version_info >= (3, 2, 0):
    from reprlib import Repr, repr, recursive_repr, __all__
else:
    # Don't use name list here to avoid module recursive use
    __all__ = (str("Repr"), str("repr"), str("recursive_repr"))

    if sys.version_info <= (3, 1, 0):
        import __builtin__ as builtins
    else:
        import builtins
    from itertools import islice
    if sys.version_info >= (3, 0):
        try:
            from _thread import get_ident
        except ImportError:
            from _dummy_thread import get_ident
    else:
        try:
            from thread import get_ident
        except ImportError:
            from dummy_thread import get_ident

    def recursive_repr(fillvalue='...'):
        '''Decorator to make a repr function return ``fillvalue`` for a recursive
        call'''

        def decorating_function(user_function):
            repr_running = set()

            def wrapper(self):
                key = id(self), get_ident()
                if key in repr_running:
                    return fillvalue
                repr_running.add(key)
                try:
                    result = user_function(self)
                finally:
                    repr_running.discard(key)
                return result

            # Can't use functools.wraps() here because of bootstrap issues
            wrapper.__module__ = getattr(user_function, '__module__')
            wrapper.__doc__ = getattr(user_function, '__doc__')
            wrapper.__name__ = getattr(user_function, '__name__')
            wrapper.__annotations__ = getattr(user_function,
                                              '__annotations__', {})
            return wrapper

        return decorating_function

    class Repr(object):

        def __init__(self):
            self.maxlevel = 6
            self.maxtuple = 6
            self.maxlist = 6
            self.maxarray = 5
            self.maxdict = 4
            self.maxset = 6
            self.maxfrozenset = 6
            self.maxdeque = 6
            self.maxstring = 30
            self.maxlong = 40
            self.maxother = 30

        def repr(self, x):
            return self.repr1(x, self.maxlevel)

        def repr1(self, x, level):
            typename = type(x).__name__
            if ' ' in typename:
                parts = typename.split()
                typename = '_'.join(parts)
            if hasattr(self, 'repr_' + typename):
                return getattr(self, 'repr_' + typename)(x, level)
            else:
                return self.repr_instance(x, level)

        def _repr_iterable(self, x, level, left, right, maxiter, trail=''):
            n = len(x)
            if level <= 0 and n:
                s = '...'
            else:
                newlevel = level - 1
                repr1 = self.repr1
                pieces = [repr1(elem, newlevel) for elem in islice(x, maxiter)]
                if n > maxiter:
                    pieces.append('...')
                s = ', '.join(pieces)
                if n == 1 and trail:
                    right = trail + right
            return '%s%s%s' % (left, s, right)

        def repr_tuple(self, x, level):
            return self._repr_iterable(x, level, '(', ')', self.maxtuple, ',')

        def repr_list(self, x, level):
            return self._repr_iterable(x, level, '[', ']', self.maxlist)

        def repr_array(self, x, level):
            header = "array('%s', [" % x.typecode
            return self._repr_iterable(x, level, header, '])', self.maxarray)

        def repr_set(self, x, level):
            x = _possibly_sorted(x)
            return self._repr_iterable(x, level, 'set([', '])', self.maxset)

        def repr_frozenset(self, x, level):
            x = _possibly_sorted(x)
            return self._repr_iterable(x, level, 'frozenset([', '])',
                                       self.maxfrozenset)

        def repr_deque(self, x, level):
            return self._repr_iterable(x, level, 'deque([', '])',
                                       self.maxdeque)

        def repr_dict(self, x, level):
            n = len(x)
            if n == 0:
                return '{}'
            if level <= 0:
                return '{...}'
            newlevel = level - 1
            repr1 = self.repr1
            pieces = []
            for key in islice(_possibly_sorted(x), self.maxdict):
                keyrepr = repr1(key, newlevel)
                valrepr = repr1(x[key], newlevel)
                pieces.append('%s: %s' % (keyrepr, valrepr))
            if n > self.maxdict:
                pieces.append('...')
            s = ', '.join(pieces)
            return '{%s}' % (s,)

        def repr_str(self, x, level):
            s = builtins.repr(x[:self.maxstring])
            if len(s) > self.maxstring:
                i = max(0, (self.maxstring-3)//2)
                j = max(0, self.maxstring-3-i)
                s = builtins.repr(x[:i] + x[len(x)-j:])
                s = s[:i] + '...' + s[len(s)-j:]
            return s

        def repr_int(self, x, level):
            s = builtins.repr(x)  # XXX Hope this isn't too slow...
            if len(s) > self.maxlong:
                i = max(0, (self.maxlong-3)//2)
                j = max(0, self.maxlong-3-i)
                s = s[:i] + '...' + s[len(s)-j:]
            return s

        def repr_instance(self, x, level):
            try:
                s = builtins.repr(x)
                # Bugs in x.__repr__() can cause arbitrary
                # exceptions -- then make up something
            except Exception:
                return '<%s instance at %x>' % (x.__class__.__name__, id(x))
            if len(s) > self.maxother:
                i = max(0, (self.maxother-3)//2)
                j = max(0, self.maxother-3-i)
                s = s[:i] + '...' + s[len(s)-j:]
            return s

    def _possibly_sorted(x):
        # Since not all sequences of items can be sorted and comparison
        # functions may raise arbitrary exceptions, return an unsorted
        # sequence in that case.
        try:
            return sorted(x)
        except Exception:
            return list(x)

    aRepr = Repr()
    repr = aRepr.repr



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#----------------------------------------------------------------------
# xoutil.html
#----------------------------------------------------------------------
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created on 2013-04-18

'''This module defines utilities to manipulate HTML.

This module backports several utilities from Python 3.2.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)

from xoutil.six import PY3 as _py3k, iteritems as iteritems_
from xoutil.string import safe_decode


if _py3k:
    from html import entities
    from html import parser
else:
    import htmlentitydefs as entities
    import HTMLParser as parser


entities.entitydefs_unicode = {}
entities.entitydefs_utf8 = {}

for name, entity in iteritems_(entities.entitydefs):
    text = entities.entitydefs_unicode[name] = safe_decode(entity, 'latin-1')
    entities.entitydefs_utf8[name] = text.encode('utf-8')
del name, entity, safe_decode, iteritems_


def _further_escape(s):
    import re
    from xoutil.string import safe_encode
    ASCII = getattr(re, 'ASCII', 0)  # Py3k
    what = re.compile(br'[\x00-\x1F\x80-\xFF]', ASCII)
    res, pos = b'', 0
    for match in what.finditer(s):
        char, start, end = match.group(), match.start(), match.end()
        assert start + 1 == end
        res += s[pos:start]
        res += b'&#' + safe_encode(str(ord(char))) + b';'
        pos = end
    res += s[pos:]
    return res

import sys
_py32 = sys.version_info >= (3, 2)
del sys

if not _py32:
    # The following is a modified copy from the Python 3.2 standard library, to
    # make xoutil forwards compatible. The modification is needed to cope with
    # the bytes/unicode issues in Python 2.7
    _escape_map = {ord('&'): '&amp;', ord('<'): '&lt;', ord('>'): '&gt;'}
    _escape_map_full = {ord('&'): '&amp;', ord('<'): '&lt;', ord('>'): '&gt;',
                        ord('"'): '&quot;', ord('\''): '&#x27;'}

    # NB: this is a candidate for a bytes/string polymorphic interface

    def escape(s, quote=True):
        """Replace special characters "&", "<" and ">" to HTML-safe sequences

        If the optional flag quote is true (the default), the quotation mark
        characters, both double quote (") and single quote (') characters are
        also translated.

        """
        from xoutil.six import text_type
        from xoutil.string import safe_decode, safe_encode
        if not isinstance(s, text_type):
            arg = safe_decode(s)
        else:
            arg = s
        if quote:
            res = arg.translate(_escape_map_full)
        else:
            res = arg.translate(_escape_map)
        if not isinstance(res, type(s)):
            return safe_encode(res)
        return res

else:
    from html import escape


del _py3k, _py32


__all__ = (str('entities'), str('parser'), str('escape'))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#----------------------------------------------------------------------
# xoutil.aop
#----------------------------------------------------------------------
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# Copyright (c) 2012 Medardo Rodríguez
# All rights reserved.
#
# Contributors: see CONTRIBUTORS and HISTORY file
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on Apr 26, 2012

'''
Very simple AOP implementation allowing method replacing in objects with change
function, reset an object to its original state and user a special super
function inner new functions used to inject the new behavior.

Aspect-oriented programming (AOP) increase modularity by allowing the separation
of cross-cutting concerns.

An aspect can alter the behavior of the base code (the non-aspect part of a
program) by applying advice (additional behavior) at various join-points (points
in a program) specified in a quantification or query called a point-cut (that
detects whether a given join point matches).

An aspect can also make structural changes to other classes, like adding members
or parents.
'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)

from .basic import complementor, inject_dependencies, weaved

__all__ = (b'complementor',
           b'inject_dependencies',
           b'weaved', )



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#----------------------------------------------------------------------
# xoutil.deprecation
#----------------------------------------------------------------------
# Copyright (c) 2012, 2013, 2014 Merchise Autrement and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on Feb 15, 2012

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)

import types
import warnings

from functools import wraps
from xoutil.six import class_types as _class_types


DEFAULT_MSG = ('{funcname} is now deprecated and it will be '
               'removed{in_version}. Use {replacement} instead.')


class DeprecationError(Exception):
    pass


# WARNING!!! Don't make deprecated depends upon anything more than six.
def deprecated(replacement, msg=DEFAULT_MSG, deprecated_module=None,
               removed_in_version=None, check_version=False):
    '''Small decorator for deprecated functions.

    Usage::

        @deprecate(new_function)
        def deprecated_function(...):
            ...

    :param replacement: Either a string or the object that replaces the
       deprecated.

    :param msg: A deprecation warning message template.  You should provide
       keyword arguments for the :func:`format` function.  Currently we pass
       the current keyword arguments: `replacement` (after some processing),
       `funcname` with the name of the currently deprecated object and
       `in_version` with the version this object is going to be removed if
       `removed_in_version` argument is not None.

       Defaults to: "{funcname} is now deprecated and it will be
       removed{in_version}. Use {replacement} instead."

    :param removed_in_version: The version the deprecated object is going to be
       removed.

    :param check_version: If True and `removed_in_version` is not None, then
       declarations of obseleted objects will raise a DeprecationError. This
       helps the release manager to keep the release clean.

       .. note:: Currently only works with setuptools' installed distributions.

    :param deprecated_module: If provided, the name of the module the
       deprecated object resides. Not needed if the deprecated object is a
       function or class.

    .. versionchanged:: 1.4.1 Introduces removed_in_version and check_version.

    '''
    from xoutil.names import nameof

    def raise_if_deprecated(target, target_version):
        import pkg_resources
        from xoutil.six import string_types
        pkg = nameof(target, inner=True, typed=True, full=True)
        pkg, _obj = pkg.rsplit('.', 1)
        dist = None
        while not dist and pkg:
            try:
                dist = pkg_resources.get_distribution(pkg)
            except pkg_resources.DistributionNotFound:
                dist = None
                if '.' in pkg:
                    pkg, _obj = pkg.rsplit('.', 1)
                else:
                    pkg, _obj = None, None
        assert dist
        if isinstance(target_version, string_types):
            target_version = pkg_resources.parse_version(target_version)
        if dist.parsed_version >= target_version:
            msg = ('A deprecated feature %r was scheduled to be '
                   'removed in version %r and it is still '
                   'alive in %r!' % (nameof(target, inner=True, full=True),
                                     str(removed_in_version),
                                     str(dist.version)))
            raise DeprecationError(msg)

    def decorator(target):
        if deprecated_module:
            funcname = deprecated_module + '.' + target.__name__
        else:
            funcname = target.__name__
        if isinstance(replacement, _class_types + (types.FunctionType, )):
            repl_name = replacement.__module__ + '.' + replacement.__name__
        else:
            repl_name = replacement
        if removed_in_version:
            in_version = ' in version ' + removed_in_version
        else:
            in_version = ''
        if isinstance(target, _class_types):
            def new(*args, **kwargs):
                if check_version and removed_in_version:
                    raise_if_deprecated(target, removed_in_version)
                warnings.warn(msg.format(funcname=funcname,
                                         replacement=repl_name,
                                         in_version=in_version),
                              stacklevel=2)
                return target.__new__(*args, **kwargs)
            # Code copied and adapted from xoutil.objects.copy_class. This is
            # done so because this module *must* not depends on any other,
            # otherwise an import cycle might be formed when deprecating a
            # class in xoutil.objects.
            from xoutil.six import iteritems as iteritems_
            from xoutil.types import MemberDescriptorType
            meta = type(target)
            attrs = {name: value
                     for name, value in iteritems_(target.__dict__)
                     if name not in ('__class__', '__mro__',
                                     '__name__', '__weakref__', '__dict__')
                     # Must remove member descriptors, otherwise the old's
                     # class descriptor will override those that must be
                     # created here.
                     if not isinstance(value, MemberDescriptorType)}
            attrs.update(__new__=new)
            result = meta(target.__name__, target.__bases__, attrs)
            return result
        else:
            @wraps(target)
            def inner(*args, **kw):
                if check_version and removed_in_version:
                    raise_if_deprecated(target, removed_in_version)
                warnings.warn(msg.format(funcname=funcname,
                                         replacement=repl_name,
                                         in_version=in_version),
                              stacklevel=2)
                return target(*args, **kw)
            return inner
    return decorator


def inject_deprecated(funcnames, source, target=None):
    '''Takes a sequence of function names `funcnames` which reside in the
    `source` module and injects them into `target` marked as deprecated. If
    `target` is None then we inject the functions into the locals of the
    calling code. It's expected it's a module.

    This function is provided for easing the deprecation of whole modules and
    should not be used to do otherwise.
    '''
    if not target:
        import sys
        frame = sys._getframe(1)
        try:
            target_locals = frame.f_locals
        finally:
            # As recommended to avoid memory leaks
            del frame
    else:
        pass
    for targetname in funcnames:
        unset = object()
        target = getattr(source, targetname, unset)
        if target is not unset:
            testclasses = (types.FunctionType, types.LambdaType) + _class_types
            if isinstance(target, testclasses):
                replacement = source.__name__ + '.' + targetname
                module_name = target_locals.get('__name__', None)
                target_locals[targetname] = deprecated(replacement,
                                                       DEFAULT_MSG,
                                                       module_name)(target)
            else:
                target_locals[targetname] = target
        else:
            warnings.warn('{targetname} was expected to be in {source}'.
                          format(targetname=targetname,
                                 source=source.__name__), stacklevel=2)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import os, sys
from setuptools import setup, find_packages

# Import the version from the release module
project_name = 'xoutil'
_current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(_current_dir, project_name))
from release import VERSION as version
from release import RELEASE_TAG

if RELEASE_TAG != '':
    dev_classifier = 'Development Status :: 4 - Beta'
else:
    dev_classifier = 'Development Status :: 5 - Production/Stable'

setup(name=project_name,
      version=version,
      description=("Collection of usefull algorithms and other very "
                   "disparate stuff"),
      long_description=open(
          os.path.join(_current_dir, 'docs', 'readme.txt')).read(),
      classifiers=[
        # Get from http://pypi.python.org/pypi?%3Aaction=list_classifiers
          dev_classifier,
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
      ],
      keywords='',
      author='Medardo Rodríguez',
      author_email='merchise.h8@gmail.com',
      url='http://www.merchise.org/',
      license='GPLv3+',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
      ],
      entry_points="""
      """,
      )



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#----------------------------------------------------------------------
# xoutil.names
#----------------------------------------------------------------------
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created on 15 avr. 2013

'''A protocol to obtain or manage object names.'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_import)


try:
    str_base = basestring
except NameError:
    str_base = str


def _key_for_value(target, value, strict=True):
    '''Returns the key that has the "value" in dictionary "target".

    if strict is True, then look first for the same object::
        >>> from functools import partial
        >>> x = {1}
        >>> y = {1}
        >>> search = partial(_key_for_value, {'x': x, 'y': y})
        >>> search(x) == search(y)
        False
        >>> search(x, strict=False) == search(y, strict=False)
        True

    This is mainly intended to find object names in stack frame variables.

    '''
    keys = list(target)     # Get keys
    i, found, equal = 0, False, None
    while (i < len(keys)) and not found:
        key = keys[i]
        item = target[key]
        if item is value:
            found = key
        elif item == value:
            if strict:
                equal = key
                i += 1
            else:
                found = key
        else:
            i += 1
    return found or equal


def module_name(target):
    if target is None:
        target = ''
    elif isinstance(target, str_base):
        res = target
    else:
        res = getattr(target, '__module__', None)
        if res is None:
            res = getattr(type(target), '__module__', '')
    if res.startswith('__') or (res in ('builtins', '<module>')):
        res = ''
    return str(res)


def nameof(target, depth=1, inner=False, typed=False, full=False):
    '''Gets the name of an object.

    .. versionadded:: 1.4.0


    The name of an object is normally the variable name in the calling stack::

        >>> from xoutil.collections import OrderedDict as sorted_dict
        >>> nameof(sorted_dict)
        'sorted_dict'

    If the `inner` flag is true, then the name is found by introspection
    first::

        >>> nameof(sorted_dict, inner=True)
        'OrderedDict'

    If the `typed` flag is true, returns the name of the type unless `target`
    is already a type or it has a "__name__" attribute, but the "__name__" is
    used only if `inner` is True.

        >>> sd = sorted_dict(x=1, y=2)
        >>> nameof(sd)
        'sd'

        >>> nameof(sd, typed=True)
        'sorted_dict'

        >>> nameof(sd, inner=True, typed=True)
        'OrderedDict'

    If `target` is an instance of a simple type (strings or numbers) and
    `inner` is true, then the name is the standard representation of `target`::

        >>> s = 'foobar'
        >>> nameof(s)
        's'

        >>> nameof(s, inner=True)
        'foobar'

        >>> i = 1
        >>> nameof(i)
        'i'

        >>> nameof(i, inner=True)
        '1'

        >>> nameof(i, typed=True)
        'int'

    If `target` isn't an instance of a simple type (strings or numbers) and
    `inner` is true, then the id of the object is used::

        >>> hex(id(sd)) in nameof(sd, inner=True)
        True

    If `full` is True, then the module where the name if defined is
    prefixed. Examples::

        >>> nameof(sd, full=True)
        'xoutil.names.sd'

        >>> nameof(sd, typed=True, full=True)
        'xoutil.names.sorted_dict'

        >>> nameof(sd, inner=True, typed=True, full=True)  # doctest: +ELLIPSIS
        '...collections.OrderedDict'

    :param depth: Amount of stack levels to skip if needed.

    '''
    from numbers import Number
    TYPED_NAME = '__name__'
    if typed and not hasattr(target, TYPED_NAME):
        target = type(target)
    if inner:
        res = getattr(target, TYPED_NAME, False)
        if res:
            if full:
                head = module_name(target)
                if head:
                    res = '.'.join((head, res))
            return str(res)
        elif isinstance(target, (str_base, Number)):
            return str(target)
        else:
            type_name = nameof(target, inner=True, typed=True, full=full)
            return str('@'.join((type_name, hex(id(target)))))
    else:
        import sys
        sf = sys._getframe(depth)
        try:
            res = False
            i, LIMIT = 0, 5   # Limit number of stack to recurse
            def getter(src):
                key = _key_for_value(l, target)
                if key and full:
                    head = src.get('__name__')
                    if not head:
                        head = sf.f_code.co_name
                    head = module_name(head)
                    if not head:
                        head = module_name(target) or None
                else:
                    head = None
                return key, head
            while not res and sf and (i < LIMIT):
                l = sf.f_locals
                key, head = getter(l)
                if not key:
                    g = sf.f_globals
                    if l is not g:
                        key, head = getter(g)
                if key:
                    res = key
                else:
                    sf = sf.f_back
                    i =+ 1
        finally:
            # TODO: on "del sf" Python says "SyntaxError: can not delete
            # variable 'sf' referenced in nested scope".
            sf = None
        if res:
            return str('.'.join((head, res)) if head else res)
        else:
            return nameof(target, depth=depth+1, inner=True, full=full)


class namelist(list):
    '''Similar to list, but only intended for storing object names.

    Constructors:

    * namelist() -> new empty list
    * namelist(collection) -> new list initialized from collection's items
    * namelist(item, ...) -> new list initialized from severals items

    Instances can be used as decorators to store names of module items
    (functions or classes)::

        >>> __all__ = namelist()
        >>> @__all__
        ... def foobar(*args, **kwargs):
        ...     'Automatically added to this module "__all__" names.'

        >>> 'foobar' in __all__
        True

    '''
    def __init__(self, *args):
        if len(args) == 1:
            from types import GeneratorType as gtype
            if isinstance(args[0], (tuple, list, set, frozenset, gtype)):
                args = args[0]
        super(namelist, self).__init__(nameof(arg, depth=2) for arg in args)

    def __add__(self, other):
        other = [nameof(item, depth=2) for item in other]
        return super(namelist, self).__add__(other)
    __iadd__ = __add__

    def __contains__(self, target):
        return super(namelist, self).__contains__(nameof(target, depth=2))

    def append(self, value):
        '''l.append(value) -- append a name object to end'''
        super(namelist, self).append(nameof(value, depth=2))
        return value    # What allow to use its instances as a decorator
    __call__ = append

    def extend(self, items):
        '''l.extend(items) -- extend list by appending items from the iterable
        '''
        items = (nameof(item, depth=2) for item in items)
        return super(namelist, self).extend(items)

    def index(self, value, *args):
        '''l.index(value, [start, [stop]]) -> int -- return first index of name

        Raises ValueError if the name is not present.

        '''
        return super(namelist, self).index(nameof(value, depth=2), *args)

    def insert(self, index, value):
        '''l.insert(index, value) -- insert object before index
        '''
        return super(namelist, self).insert(index, nameof(value, depth=2))

    def remove(self, value):
        '''l.remove(value) -- remove first occurrence of value

        Raises ValueError if the value is not present.

        '''
        return list.remove(self, nameof(value, depth=2))


class strlist(list):
    '''Similar to list, but only intended for storing ``str`` instances.

    Constructors:
        * strlist() -> new empty list
        * strlist(collection) -> new list initialized from collection's items
        * strlist(item, ...) -> new list initialized from severals items

    Last versions of Python 2.x has a feature to use unicode as standard
    strings, but some object names can be only ``str``. To be compatible with
    Python 3.x in an easy way, use this list.
    '''
    def __init__(self, *args):
        if len(args) == 1:
            from types import GeneratorType as gtype
            if isinstance(args[0], (tuple, list, set, frozenset, gtype)):
                args = args[0]
        super(strlist, self).__init__(str(arg) for arg in args)

    def __add__(self, other):
        other = [str(item) for item in other]
        return super(strlist, self).__add__(other)
    __iadd__ = __add__

    def __contains__(self, target):
        return super(strlist, self).__contains__(str(target))

    def append(self, value):
        '''l.append(value) -- append a name object to end'''
        super(strlist, self).append(str(value))
        return value    # What allow to use its instances as a decorator
    __call__ = append

    def extend(self, items):
        '''l.extend(items) -- extend list by appending items from the iterable
        '''
        items = (str(item) for item in items)
        return super(strlist, self).extend(items)

    def index(self, value, *args):
        '''l.index(value, [start, [stop]]) -> int -- return first index of name

        Raises ValueError if the name is not present.

        '''
        return super(strlist, self).index(str(value), *args)

    def insert(self, index, value):
        '''l.insert(index, value) -- insert object before index
        '''
        return super(strlist, self).insert(index, str(value))

    def remove(self, value):
        '''l.remove(value) -- remove first occurrence of value

        Raises ValueError if the value is not present.

        '''
        return list.remove(self, str(value))


__all__ = strlist('nameof', 'namelist', 'strlist')



# -*- coding: utf-8 -*-
#----------------------------------------------------------------------
# xoutil.compat
#----------------------------------------------------------------------
#
# ============================ Original copyright notice ===================
# Copyright (C) 2005-2011, 2014 the SQLAlchemy authors and contributors
#
# This module is based on part of SQLAlchemy and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php
# ============================ Original copyright notice ===================

"""Handle Python version/platform incompatibilities."""

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)

import warnings
warnings.warn('The xoutil.compat module is being deprecated in favor of '
              'xoutil.six')
import sys

# TODO: Changes in Python 3
#    * remove "iteritems", "itervalues", "iterkeys" for dict
#    * basestring, str, unicode
#    * "__builtin__" is now "builtins"
#    * "import cPickle as pickle" not needed any more
#    * __metaclass__
#    * "m.im_class" is "m.__self__.__class__"
#    * many "types.*" are removed: types.ClassType = type
#    * "StandardError" removed
#    * xrange
#    * "long" integrated with "int"
#    * The StringIO and cStringIO modules are gone. Instead, import the io
#      module and use io.StringIO or io.BytesIO for text and data respectively.
#    * "exec" is a function
#    * Integer literals no longer support a trailing l or L.
#    * String literals no longer support a leading u or U.
#    * Classic classes are gone
#    * except type, var is not longer supported
#    * "__nonzero__" -> "__bool__"

# TODO: Make compatible all libraries and applications with Python 2.7

try:
    import threading
except ImportError:
    import dummy_threading as threading

py32 = sys.version_info >= (3, 2)
py33 = sys.version_info >= (3, 3)
py3k = getattr(sys, 'py3kwarning', False) or sys.version_info >= (3, 0)
jython = sys.platform.startswith('java')
pypy = hasattr(sys, 'pypy_version_info')
win32 = sys.platform.startswith('win')

if py3k:
    import builtins
    str_base = str
    str_types = (str, )
    u = _unicode = str
    ext_str_types = (bytes, str)
    TypeType = type
    class_types = (type, )
    integer = long = int
    integers = (int, )
    # TODO: [manu] Remove all "xrange" and all aliases:
    #       "xoutil.compat.range" will be "xrange" in Python 2.x and will
    #       remains unchanged in Python 2.0
    from builtins import range as xrange
    xrange_ = xrange
    range_ = range = lambda *args: list(xrange(*args))
    exec_ = getattr(builtins, "exec")
    def execfile_(fname, *args):
        return exec_(compile(open(fname, 'rb').read(), fname, 'exec'), *args)
else:
    str_base = basestring
    str_types = (str, unicode)
    _unicode = unicode
    ext_str_types = (str, unicode)
    from types import TypeType
    class_types = (type, TypeType)
    # FIXME: [manu] In Py2  isinstance(1, long) is False.
    integers = (long, int)
    integer = long
    from __builtin__ import xrange, range, execfile
    xrange_ = xrange
    range_ = range
    execfile_ = execfile

    def exec_(code, globs=None, locs=None):
        """Execute code in a namespace."""
        if globs is None:
            frame = sys._getframe(1)
            globs = frame.f_globals
            if locs is None:
                locs = frame.f_locals
            del frame
        elif locs is None:
            locs = globs
        exec("""exec code in globs, locs""")


if py3k:
    set_types = set
elif sys.version_info < (2, 6):
    import sets
    set_types = set, sets.Set
else:
    # 2.6 deprecates sets.Set, but we still need to be able to detect them
    # in user code and as return values from DB-APIs
    ignore = ('ignore', None, DeprecationWarning, None, 0)
    import warnings
    try:
        warnings.filters.insert(0, ignore)
    except Exception:
        import sets
    else:
        import sets
        warnings.filters.remove(ignore)
    set_types = set, sets.Set

if py3k:
    import pickle
else:
    try:
        import cPickle as pickle
    except ImportError:
        import pickle


if py3k:
    import configparser
else:
    import ConfigParser as configparser
    ConfigParser = configparser


if py3k:
    from inspect import getfullargspec as inspect_getfullargspec
else:
    from inspect import getargspec as inspect_getfullargspec


if py3k and not py32:
    def callable(obj):
        '''callable(object) -> bool

        Return whether the object is callable (i.e., some kind of function).
        Note that classes are callable, as are instances of classes with a
        __call__() method.

        '''
        return hasattr(obj, '__call__')
else:
    # Removed in Python 3 and brought back in 3.2.  brilliant!
    try:
        from builtins import callable
    except ImportError:
        from __builtin__ import callable


if py3k:
    def cmp(a, b):
        return (a > b) - (a < b)

    # Remove this import, always use it directly
    from functools import reduce
else:
    from __builtin__ import cmp, reduce

try:
    from collections import defaultdict
except ImportError:
    class defaultdict(dict):
        def __init__(self, default_factory=None, *a, **kw):
            if (default_factory is not None and
                not hasattr(default_factory, '__call__')):
                raise TypeError('first argument must be callable')
            dict.__init__(self, *a, **kw)
            self.default_factory = default_factory
        def __getitem__(self, key):
            try:
                return dict.__getitem__(self, key)
            except KeyError:
                return self.__missing__(key)
        def __missing__(self, key):
            if self.default_factory is None:
                raise KeyError(key)
            self[key] = value = self.default_factory()
            return value
        def __reduce__(self):
            if self.default_factory is None:
                args = tuple()
            else:
                args = self.default_factory,
            return type(self), args, None, None, self.iteritems()
        def copy(self):
            return self.__copy__()
        def __copy__(self):
            return type(self)(self.default_factory, self)
        def __deepcopy__(self, memo):
            import copy
            return type(self)(self.default_factory,
                              copy.deepcopy(self.items()))
        def __repr__(self):
            return 'defaultdict(%s, %s)' % (self.default_factory,
                                            dict.__repr__(self))


# find or create a dict implementation that supports __missing__
class _probe(dict):
    def __missing__(self, key):
        return 1

try:
    _probe()['missing']
    py25_dict = dict
except KeyError:
    class py25_dict(dict):
        def __getitem__(self, key):
            try:
                return dict.__getitem__(self, key)
            except KeyError:
                try:
                    missing = self.__missing__
                except AttributeError:
                    raise KeyError(key)
                else:
                    return missing(key)
finally:
    del _probe


try:
    import hashlib
    _md5 = hashlib.md5
except ImportError:
    import md5
    _md5 = md5.new


def md5_hex(x):
    # Py3K
    #x = x.encode('utf-8')
    m = _md5()
    m.update(x)
    return m.hexdigest()

import time
if win32 or jython:
    time_func = time.clock
else:
    time_func = time.time

if sys.version_info >= (2, 5):
    def decode_slice(slc):
        """decode a slice object as sent to __getitem__.

        takes into account the 2.5 __index__() method, basically.

        """
        ret = []
        for x in slc.start, slc.stop, slc.step:
            if hasattr(x, '__index__'):
                x = x.__index__()
            ret.append(x)
        return tuple(ret)
else:
    def decode_slice(slc):
        return (slc.start, slc.stop, slc.step)

if sys.version_info >= (2, 6):
    from operator import attrgetter as dottedgetter
else:
    def dottedgetter(attr):
        def g(obj):
            for name in attr.split("."):
                obj = getattr(obj, name)
            return obj
        return g

import decimal  # TODO: Why 'decimal' is here?


if py3k:    # pragma: no cover
    def iteritems_(d):
        return d.items()
    def itervalues_(d):
        return d.values()
    def iterkeys_(d):
        return d.keys()
else:
    def iteritems_(d):
        return d.iteritems()
    def itervalues_(d):
        return d.itervalues()
    def iterkeys_(d):
        return d.iterkeys()

try:
    import ConfigParser as configparser
except:
    import configparser     # Name changed in Python3

try:
    from future_builtins import zip
except ImportError:
    from builtins import zip
izip = zip

try:
    from future_builtins import map
except ImportError:
    from builtins import map
imap = map

try:
    from itertools import izip_longest as zip_longest
except ImportError:
    from itertools import zip_longest
izip_longest = zip_longest


if py3k:
    from builtins import chr
else:
    from __builtin__ import unichr as chr


if py3k:
    from builtins import input
else:
    from __builtin__ import raw_input as input



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#----------------------------------------------------------------------
# tests.test_iterators
#----------------------------------------------------------------------
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created on 2013-04-12

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)



def test_first_n_no_filling():
    from xoutil.iterators import first_n
    try:
        next(first_n((), 1))
        assert False, 'Should have raised an StopIteration'
    except StopIteration:
        pass
    except:
        assert False, 'Should have raised an StopIteration'

def test_first_n_filling_by_cycling():
    from xoutil.iterators import first_n
    assert list(first_n((), 10, range(5))) == [0, 1, 2, 3, 4] * 2


def test_first_n_repeat_filling_by_repeating():
    from xoutil.iterators import first_n
    from itertools import repeat
    assert list(first_n((), 10, '0')) == list(repeat('0', 10))


def test_first_n_simple():
    from xoutil.iterators import first_n
    assert list(first_n(range(100), 10, 0)) == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]


def test_slides():
    from xoutil.iterators import slides
    assert list(slides(range(1, 11))) == [(1, 2),
                                          (3, 4),
                                          (5, 6),
                                          (7, 8),
                                          (9, 10)]

def test_slides_filling():
    from xoutil.iterators import slides
    assert list(slides(range(1, 5), 3)) == [(1, 2, 3), (4, None, None)]


def test_slides_with_repeating_filling():
    from xoutil.iterators import slides
    assert list(slides(range(1, 11), width=3, fill=None)) == [(1, 2, 3), (4, 5, 6), (7, 8, 9), (10, None, None)]

def test_slides_with_cycling_filling():
    from xoutil.iterators import slides
    assert list(slides(range(1, 12), width=5, fill=(1, 2))) == [(1, 2, 3, 4, 5), (6, 7, 8, 9, 10), (11, 1, 2, 1, 2)]


def test_continuously_slides():
    from xoutil.iterators import continuously_slides
    trigrams = list(''.join(x) for x in continuously_slides('maupasant', 3, ''))
    assert 'mau' in trigrams
    assert 'aup' in trigrams
    assert 'upa' in trigrams
    assert 'pas' in trigrams
    assert 'asa' in trigrams
    assert 'san' in trigrams
    assert 'ant' in trigrams
    assert len(trigrams) == 7



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#----------------------------------------------------------------------
# xoutil.aop.basic
#----------------------------------------------------------------------
# Copyright (c) 2012, 2013, 2014 Merchise Autrement and Contributors
# All rights reserved.
#
# Author: Medardo Rodríguez
# Contributors: see CONTRIBUTORS and HISTORY file
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on Mar 23, 2012

'''
Very simple AOP implementation allowing method replacing in objects with
change function, reset an object to its original state and user a special
super function inner new functions used to inject the new behavior.

Aspect-oriented programming (AOP) increase modularity by allowing the
separation of cross-cutting concerns.

An aspect can alter the behavior of the base code (the non-aspect part of a
program) by applying advice (additional behavior) at various join-points
(points in a program) specified in a quantification or query called a point-
cut (that detects whether a given join point matches).

An aspect can also make structural changes to other classes, like adding
members or parents.
'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)

from types import MethodType, FunctionType
from contextlib import contextmanager

from xoutil.deprecation import deprecated as _deprecated

from xoutil.names import strlist as strs
__all__ = strs('complementor', 'contextualized', 'inject_dependencies',
               'weaved')
del strs


def _update(attrs, *sources):
    '''
    Updates attrs from sources:

    - Every source with a __name__ that is not a class is updated into attrs.

    - For every source that is a class, it's public attributes and all methods
      are updated into attrs.
    '''
    from xoutil.six import class_types, iteritems as iteritems_
    attrs.update({str(f.__name__): f for f in sources
                  if (not isinstance(f, class_types) and
                      getattr(f, '__name__', False))})

    attrs.update({str(name): member
                  for cls in sources if isinstance(cls, class_types)
                  # XXX: [manu] Use cls.__dict__ instead of xdir(cls) since
                  # getattr(cls, attr) would not yield the classmethod and
                  # staticmethod wrappers.
                  for name, member in iteritems_(cls.__dict__)
                  if not name.startswith('_') or
                     isinstance(member, FunctionType)})
    return attrs


@_deprecated('None', msg="This entire module is deprecated and will be "
             "removed.", removed_in_version='1.6.0')
def complementor(*sources, **attrs):
    '''
    Returns a decorator to be applied to a class in order to add attributes
    in a smart way:

    - if the attribute is a dictionary and exists in the decorated class, it's
      updated.

    - If a list, tuple or set, the new value is appended.

    - Methods declared in the class that are replacements are renamed to
      "_super_<name>", but the docstring and names are copied to their
      replacement.

    - All other values are replaced.

    The following code tests show each case::

        >>> def hacked_init(self, *args, **kw):
        ...    print('Hacked')
        ...    self._super___init__(*args, **kw)

        >>> from xoutil.six.moves import range
        >>> range_ = lambda *a: list(range(*a))
        >>> @complementor(somedict={'a': 1, 'b': 2}, somelist=range_(5, 10),
        ...               __init__=hacked_init)
        ... class Someclass(object):
        ...     somedict = {'a': 0}
        ...     somelist = range_(5)
        ...
        ...     def __init__(self, d=None, l=None):
        ...         'My docstring'
        ...         if d:
        ...             self.somedict.update(d)
        ...         if l:
        ...             self.somelist.extend(l)

        # It's best to do comparison with dicts since key order may not be
        # preserved.
        >>> Someclass.somedict == {'a': 1, 'b': 2}
        True

        >>> Someclass.somelist
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

        >>> instance = Someclass(d={'c': 12}, l=(10, ))
        Hacked

        >>> instance.somedict == {'a': 1, 'b': 2, 'c': 12}
        True

        >>> instance.somelist
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        >>> Someclass.__init__.__doc__
        'My docstring'

    :param sources: If any positional arguments `sources` are given then for
                    each of them:

                    - If it's not a class (an instance of `type`) and it has a
                      `__name__`, it will be updated into `attrs` and treated
                      according to the rules before.

                    - If it's a class all it's public non-method attributes
                      are updated into `attrs`, and all it's methods (public
                      or private) are updated as well.

    Notice the order in which this is done: class takes precedence over other
    kind of sources, and sources takes precedence over keyword arguments.
    '''
    def inner(cls):
        from collections import (Mapping, MutableMapping,
                                 MutableSequence as List,
                                 Set)
        from xoutil import Unset
        from xoutil.six import iteritems as iteritems_
        for attr, value in iteritems_(attrs):
            attr = str(attr)
            assigned = attr in cls.__dict__
            if assigned:
                ok = isinstance
                # XXX: [manu] In order to be Python 2 and 3 compatible is best
                # to get things from the class' dictionary, then current would
                # be a function for methods, a classmethod for classmethods and
                # a staticmethod for staticmethods.
                current = cls.__dict__.get(attr, None)
                if ok(value, Mapping) and ok(current, MutableMapping):
                    current.update(value)
                    value = Unset
                elif ok(value, List) and ok(current, List):
                    current.extend(value)
                    value = Unset
                elif ok(value, tuple) and ok(current, tuple):
                    value = current + value
                elif ok(value, Set) and ok(current, Set):
                    value = current | value
                elif ok(value, (FunctionType, classmethod, staticmethod)) and current:
                    setattr(cls, str('_super_%s') % attr, current)
            else:
                current = None
            if value is not Unset:
                if current and not getattr(value, '__doc__', False):
                    from functools import update_wrapper
                    update_wrapper(value, current)
                setattr(cls, attr, value)
        return cls

    _update(attrs, *sources)
    return inner


@_deprecated('None', msg="This entire module is deprecated and will be "
             "removed.", removed_in_version='1.6.0')
def contextualized(context, *sources, **attrs):
    '''
    Another decorator very similar to :func:`complementor`, but this wraps
    every method injected inside the context provided by `context`.

        >>> from xoutil.context import context
        >>> class FooBazer(object):
        ...    def inside(self):
        ...        if context['in-context']:
        ...            print('in-context')
        ...        else:
        ...            print('outside-context')

        >>> @contextualized(context('in-context'), FooBazer)
        ... class Foobar(object):
        ...    def outside(self):
        ...        print('not-context')

        >>> foo = Foobar()
        >>> foo.inside()
        in-context

        >>> foo.outside()
        not-context

    '''
    def wrap(method):
        def inner(self, *args, **kwargs):
            with context:
                if getattr(method, 'im_self', None):
                    return method(*args, **kwargs)
                else:
                    return method(self, *args, **kwargs)
        return inner

    _update(attrs, *sources)
    wrapped_attrs = {}
    for name, value in attrs.items():
        if isinstance(value, (FunctionType, MethodType)):
            wrapped_attrs[name] = wrap(value)
    return complementor(**wrapped_attrs)


@_deprecated('None', msg="This entire module is deprecated and will be "
             "removed.", removed_in_version='1.6.0')
def inject_dependencies(target, *sources, **attrs):
    'Injects/replaces the sources for the given target.'
    cls = target if isinstance(target, type) else target.__class__
    _merchise_extended = cls.__dict__.get('_merchise_extended',
                                          {'depth': 0}).copy()
    _merchise_extended['depth'] += 1
    attrs.update({b'__doc__': cls.__doc__, b'__module__': b'merchise.builtin',
                  '_merchise_extended': _merchise_extended})
    _update(attrs, *sources)
    extended_class = type(cls.__name__, (cls,), attrs)
    if not isinstance(target, type):
        try:
            target.__class__ = extended_class
        except:
            class wrapper(target.__class__):
                pass
            target = wrapper(target)
            target.__class__ = extended_class
        result = target
    else:
        result = extended_class
    return result


@_deprecated('None', msg="This entire module is deprecated and will be "
             "removed.", removed_in_version='1.6.0')
@contextmanager
def weaved(target, *sources, **attrs):
    '''
    Returns a context manager that weaves `target` with all the `sources` and
    the `attrs` weaved into it. This method yields the weaved object into the
    context manager, so you have access to it. Upon exiting the context
    manager, the `target` is reset to it's previous state (if possible).

    For example, in the following code::

        >>> class Test(object):
        ...    def f(self, n):
        ...        return n

        >>> test = Test()


    You may want to record each call to ``f`` on the ``test`` object::

        >>> def f(self, n):
        ...     print('Log this')
        ...     return super(_ec, self).f(n) # Notice the assignment to _ec

        >>> with weaved(test, f) as test:
        ...    _ec = test.__class__ # Needed
        ...    test.f(10)
        Log this
        10

    But once you leave the context, ``test`` will stop logging::

        >>> test.f(10)
        10

    You may nest several calls to this:

        >>> def f2(self, n):
        ...    print('Done it')
        ...    return super(_ec2, self).f(n)

        >>> with weaved(test, f) as test:
        ...    _ec = test.__class__
        ...    with weaved(test, f=f2) as test:
        ...        _ec2 = test.__class__
        ...        test.f(10)
        ...    test.f(12)
        Done it
        Log this
        10
        Log this
        12

        >>> test.f(10)
        10

    '''
    try:
        result = inject_dependencies(target, *sources, **attrs)
        yield result
    except:
        result = None
        raise
    finally:
        if result is target:
            cls = target.__class__
            res = 0
            depth = 0
            while depth == 0 and cls.__dict__.get('_merchise_extended'):
                depth = cls.__dict__.get('_merchise_extended',
                                         {}).get('depth', 0)
                cls = cls.__base__
                res += 1
            if res > 0:
                target.__class__ = cls

del _deprecated



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#----------------------------------------------------------------------
# xoutil.tests.test_metaclass
#----------------------------------------------------------------------
# Copyright (c) 2013, 2014 Merchise Autrement and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under
# the terms of the LICENCE attached in the distribution package.
#
# Created on 2013-05-06

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        unicode_literals as _py3_unicode,
                        absolute_import as _py3_abs_imports)


def test_basic_inline_metaclass():
    from xoutil.objects import metaclass

    class Meta(type):
        pass

    class Base(metaclass(Meta)):
        pass

    class Entity(Base):
        pass

    assert isinstance(Base, Meta), 'Wrong metaclass %r' % type(Base)
    assert isinstance(Entity, Meta), 'Wrong metaclass %r' % type(Entity)
    assert Entity.__base__ is Base
    assert Base.__base__ is object


def test_no_double_registration_with_inlinemetaclass():
    import sys
    from xoutil.objects import metaclass
    py32 = sys.version_info >= (3, 2)

    class RegisteringType(type):
        classes = []

        def __new__(cls, name, bases, attrs):
            res = super(RegisteringType, cls).__new__(cls, name, bases, attrs)
            cls.classes.append(res)
            return res

    class Base(metaclass(RegisteringType)):
        pass

    class SubType(RegisteringType):
        pass

    class Egg(metaclass(SubType), Base):
        pass

    assert len(RegisteringType.classes) == 2

    class Spam(Base, metaclass(SubType)):
        'Like "Egg" but registered twice in Python 2.x.'

    assert len(RegisteringType.classes) == (3 if py32 else 4)


def test_inlinemetaclass_decorator_with_slots():
    from xoutil.objects import metaclass
    from xoutil.types import MemberDescriptorType

    class Meta(type):
        pass

    class Base(metaclass(Meta)):
        __slots__ = 'attr'

    class Ok(metaclass(Meta)):
        def __init__(self, **kwargs):
            self.__dict__ = kwargs

        @classmethod
        def clmethod(cls):
            return cls

        @staticmethod
        def stmethod(echo):
            return echo

        def echo(self, echo):
            return self, echo

    assert isinstance(Base.attr, MemberDescriptorType)
    assert isinstance(Base, Meta), 'Wrong metaclass %r' % type(Base)
    assert isinstance(Ok, Meta), 'Wrong metaclass %r' % type(Ok)

    b = Base()
    b.attr = 1
    try:
        b.another = 2
        assert False, 'Should have raised AttributeError'
    except AttributeError:
        pass
    except:
        assert False, 'Should have raised AttributeError'

    ok = Ok(name='ok')
    assert ok.stmethod(ok) == ok
    assert ok.clmethod() == Ok
    assert ok.echo(1) == (ok, 1)
    assert ok.name == 'ok'


def test_prepare_a_class():
    import sys
    from xoutil.objects import metaclass

    class ClassDict(dict):
        pass

    class Meta(type):
        @classmethod
        def __prepare__(cls, name, bases, **kwargs):
            return ClassDict(kwargs, __prepared__=True)

        def __new__(cls, name, bases, attrs, **kwargs):
            assert isinstance(attrs, ClassDict), 'Wrong type %r' % type(attrs)
            assert attrs['__prepared__'] is True
            return super(Meta, cls).__new__(cls, name, bases, attrs)

    class AnotherClassDict(ClassDict):
        pass

    class Submeta(Meta):
        @classmethod
        def __prepare__(cls, name, bases, **kwargs):
            res = super(Submeta, cls).__prepare__(name, bases, **kwargs)
            return AnotherClassDict(res)

        def __new__(cls, name, bases, attrs, **kwargs):
            res = super(Submeta, cls).__new__(cls, name, bases, attrs,
                                              **kwargs)
            assert isinstance(attrs, AnotherClassDict)
            return res

    class SubmetaLight(Meta):
        pass

    class Foobar(metaclass(Meta, kwarg1='Foobar')):
        pass

    try:
        class Foobar2(Foobar):
            pass
    except AssertionError:
        if sys.version_info >= (3, 0):
            raise
    else:
        if sys.version_info < (3, 0):
            assert False, 'Should have raised'

    class Foobaz(metaclass(Submeta)):
        pass

    class Eggbag(metaclass(SubmetaLight)):
        pass
